#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""
Custom error classes + error handling and warning mechanisms.

All non-standard errors raised by lnk should be defined here.
"""

import click
import ecstasy
import googleapiclient.errors
import re
import requests
import sys

from collections import namedtuple

Message = namedtuple('Message', ['what', 'level'])

class Error(Exception):
	"""
	Exception base class for all errors.

	Implements a system of multiple verbosity levels
	such that error output can be controlled according
	to the verbosity setting specified by the user
	(how many -v are passed).

	Attributes:
		what (str): A plain string message specifying what happened.
		levels (list): A list of strings for each verbosity level.
	"""

	def __init__(self, what, **additional):
		self.what = what or 'Something bad happened.'

		#\a is the bell character (makes a 'beep' sound)
		additional['Error'] = Message('\a{0}'.format(self.what), 0)
		additional['Type'] = Message(type(self).__name__, 2)

		self.levels = self.get_levels(additional)

		super(Error, self).__init__(self.what)

	def level(self, verbosity):
		"""
		Fetches a list of all non-empty levels for a given verbosity.

		Arguments:
			verbosity (int): The verbosity level (e.g. 0 for 'what').

		Returns:
			All levels for the given verbosity, but filtering out
			empty ones.
		"""
		levels = self.levels[:verbosity + 1]

		return [i for i in levels if i]

	@staticmethod
	def get_levels(additional):
		"""
		Transforms and formats error levels.

		Each additional level passed to the constructor of
		Error is either the error string, in which case its level
		is 1 or a Error.Message tuple consisting of that same
		error string and additionally an integer specifying the
		level. This information is parsed here, such that the
		result is a list of level strings. Moreover, each
		key/value pair of each error message is formatted
		with ecstasy to make it look pretty.

		Arguments:
			additional (dict): The additional error messages.

		Returns:
			A list of strings for each level of verbosity.
		"""
		# Each nested list corresponds to one further level of verbosity
		levels = [[], [], [], []]
		for key, value in additional.items():
			if key and value:
				level = 1 # default
				if isinstance(value, Message):
					level = value.level
					value = value.what
				line = '<{0}>: {1}'.format(key, value)
				line = ecstasy.beautify(line, ecstasy.Color.Red)
				levels[level].append(line)

		return ['\n'.join(level) if level else None for level in levels]

class HTTPError(Error):
	"""
	Raised in case of a faulty HTTP request.

	Usually because the data provided by the user was ill-formed
	(such as an expanded url where a shortened one is expected).
	Additionally may have a code and status information.
	"""
	def __init__(self, what, code=None, status=None, **additional):
		super(HTTPError, self).__init__(what,
										Code=code,
										Status=status,
										**additional)

class APIError(Error):
	"""
	Raised if there was a problem communicating with the API.

	Meaning if the problem was not ill-formed data as is the case
	for an HTTPError, but something specific to the API (possibly
	associated with its key/access-token).
	"""
	def __init__(self, what, message=None, **additional):
		super(APIError, self).__init__(what, Message=message, **additional)

class UsageError(Error):
	"""
	Raised for invalid command-line usage.

	A UsageError is raised by lnk if some cli-related badness is not
	handled by click (such as misisng urls). If invalid command-line
	usage is handled by click, this is detected and click's exception
	is then re-raised as a UsageError (i.e. any faulty cli input is
	always reported as an error of this type).
	"""
	def __init__(self, what, **additional):
		super(UsageError, self).__init__(what, **additional)

class InvalidKeyError(Error):
	"""
	Raised when an invalid key is passed to the config command.
	"""
	def __init__(self, what, **additional):
		super(InvalidKeyError, self).__init__(what, **additional)

class ConnectionError(Error):
	"""
	Raised if there was a problem connecting with the API.

	(Usually if the user is not connected to the interwebs).
	"""
	def __init__(self, what, hint=None, **additional):
		hint = hint or 'Are you connected to the webs?'
		super(ConnectionError, self).__init__(what, Hint=hint, **additional)

class AuthorizationError(Error):
	"""
	Raised in case of an authorization error.

	This is mostly the case initially, when the user first runs
	lnk and has noet yet authenticated himself and authorized
	lnk to access private information (such as user data for bitly).
	"""
	def __init__(self,
				service,
				what='Missing authorization code!',
				**additional):
		logo = ecstasy.beautify('<lnk>', ecstasy.Style.Bold)
		details = 'You have not yet authorized {0} to '.format(logo)
		details += 'access your private {0} information. '.format(service)
		hint = "Run 'lnk {0} key --generate'.".format(service)
		super(AuthorizationError, self).__init__(what,
												 Details=details,
												 Hint=hint,
												 **additional)

class InternalError(Error):
	"""
	Raised when something went wrong internally.

	Will be thrown only within methods that are non-accessible via
	the API but are used forinternal features or processing.
	Basically get mad at the project creator.
	"""

	def __init__(self, what, **additional):
		"""
		Initializes the Error super-class.

		Arguments:
			what (str): A descriptive string regarding the cause of the error.
		"""
		super(InternalError, self).__init__(what, **additional)

class Catch(object):

	"""
	Exception-handling class.

	Catch is essentially a interface to its main method 'catch',
	but adds the ability to add a usage string that is displayed
	optionally in case of a UsageError, and also the service being
	used while the exception is thrown. The information about the
	url-shortening service is of use for ConnectionErrors.

	Attributes:
		verbosity (int): The verbosity level (defaults to 0).
		usage (str): Optionally, a usage string to display for UsageErrors.
		service (str): The url-shortening service currently in use.
	"""

	def __init__(self, verbosity=0, usage=None, service=None):
		self.verbosity = verbosity
		self.usage = usage
		self.service = '{0} '.format(service) if service else ''

	def catch(self, function, *args, **kwargs):
		"""
		Executes a function and handles any potential exceptions.

		Arguments:
			function (func): The function to execute.
			args (variadic): The positional arguments to pass to the function call.
			kwargs (variadic): The keyword arguments to pass to the function call.
		"""
		try:
			try:
				function(*args, **kwargs)
			# Re-raise as an error we can handle (and format)
			except click.ClickException:
				error = self.get_error()
				raise UsageError(error.message)
			except googleapiclient.errors.HttpError:
				self.handle_google_error()
			except requests.exceptions.ConnectionError:
				error = self.get_error()
				raise ConnectionError('Could not establish connection '
									  'to {0}server!'.format(self.service))
			except click.exceptions.Abort:
				click.echo() # Just the newline 
		except Error:
			self.handle_error()

	def handle_error(self):
		"""
		Handles an Error.

		Takes care of retrieveing the error message string from the
		Error that is appropriate to the verbosity level passed to
		the constructor of Catch. This message string is then printed
		to stdout. If the Error was a UsageError, the usage string
		is printed too.
		"""
		error = self.get_error()
		click.echo('\n'.join(error.level(self.verbosity)))
		if isinstance(error, UsageError) and self.usage:
			click.echo(self.usage)

	@staticmethod
	def handle_google_error():
		"""
		Handles exceptions raised by the goo.gl api client.

		Exceptions raised by the goo.gl api client do not make
		any error clean message available, but only a long, formatted
		string ready for output. This method takes care of retrieving
		the actualy 'reason' for the exception via regular expressions.
		This reason is then re-raised as an HTTPError that can be handled
		by Catch.
		"""
		error = Catch.get_error()
		match = re.search(r'<HttpError.+returned "([\w\s]+)">$',
						  str(error))
		what = '{0}.'.format(match.group(1))
		if what == 'Required.':
			what = 'Invalid link.'

		raise HTTPError(what)

	@staticmethod
	def get_error():
		"""Retrieves the last error raised in the system."""
		# type, value, traceback
		_, error, _ = sys.exc_info()

		return error

def catch(function, *args, **kwargs):
	"""
	Convenience function for a Catch object with default settings.

	Arguments:
		function (func): The function to execute.
		args (variadic): The positional arguments to pass to the function call.
		kwargs (variadic): The keyword arguments to pass to the function call.
	"""
	Catch().catch(function, *args, **kwargs)

def warn(what):
	"""
	Outputs and formats a warning.

	Arguments:
		what (str): The warning string to output.
	"""
	what = '\a<Warning>: {0}'.format(what)
	formatted = ecstasy.beautify(what, ecstasy.Color.Yellow)
	click.echo(formatted)



# -*- coding: utf-8 -*-

"""Package settings."""

__title__ = 'lnk'
__version__ = '0.1.4'
__author__ = 'Peter Goldsborough'
__email__ = 'petergoldsborough@hotmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2015 Peter Goldsborough'
__url__ = 'https://github.com/goldsborough/lnk'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import ecstasy
import os
import pytest
import random
import requests
import time

from collections import namedtuple

import tests.paths
import lnk.bitly.user

VERSION = 3
API = 'https://api-ssl.bitly.com/v{0}'.format(VERSION)
with open(os.path.join(tests.paths.TEST_PATH, 'bitly', 'token')) as source:
	ACCESS_TOKEN = source.read()


def request_info():
	response = requests.get('{0}/user/info'.format(API),
							params=dict(access_token=ACCESS_TOKEN))

	return response.json()['data']

def request_history():
	response = requests.get('{0}/user/link_history'.format(API),
							params=dict(access_token=ACCESS_TOKEN))
	data = response.json()['data']['link_history']
	history = [i['link'] for i in data]
	history.sort()

	return history


@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', [
		'user',
		'only',
		'hide',
		'sets',
		'all',
		'selected',
		'formatted',
		'template',
		'history'
		])

	user = lnk.bitly.user.User(raw=True)
	user.parameters['access_token'] = ACCESS_TOKEN
	user.history.parameters['access_token'] = ACCESS_TOKEN
	only = ['name', 'date', 'privacy']
	hide = ['login']
	sets = {
	    'name': 'full_name',
	    'date': 'member_since',
	    'privacy': 'default_link_privacy',
	}
	data = request_info()
	selected = dict((k, data[k]) for k in data if k in sets.values())
	formatted = []
	for key, value in selected.items():
		if isinstance(value, list):
			formatted.append('{0}: '.format(key.title()))
			formatted += [' + {0}'.format(t) for t in value]
		else:
			formatted.append(user.format(key, value))
	formatted.sort()
	template = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	history = request_history()

	return Fixture(user,
				   only,
				   hide,
				   sets,
				   data,
				   selected,
				   formatted,
				   template,
				   history)


def test_initializes_well(fixture):
	
	assert hasattr(fixture.user, 'history')
	assert hasattr(fixture.user, 'keys')
	assert hasattr(fixture.user, 'list_item')
	assert isinstance(fixture.user.keys, dict)
	assert 'full_name' in fixture.user.keys
	assert fixture.user.keys['full_name'] == 'Full Name'


def test_order_works(fixture):
	expected = [
		'full_name',
		'member_since',
		'default_link_privacy'
	]

	keys = list(fixture.selected.keys())
	random.shuffle(keys)
	shuffled = dict((key, fixture.selected[key]) for key in keys)
	result = fixture.user.order(shuffled, fixture.sets.values())

	assert list(result.keys()) == expected


def test_request_works(fixture):
	result = fixture.user.request(fixture.sets)
	expected = fixture.user.order(fixture.selected, fixture.sets)

	assert result == expected

def test_format_maps_keys_well(fixture):
	result = fixture.user.format('full_name', 'Satan')
	expected = 'Full Name: Satan'
	assert result == expected

	result = fixture.user.format('default_link_privacy', 'public')
	expected = 'Link privacy: public'
	assert result == expected


def test_format_handles_None_well(fixture):
	result = fixture.user.format('full_name', None)
	expected = 'Full Name: None'

	assert result == expected


def test_format_parses_times_well(fixture):
	now = time.time()
	result = fixture.user.format('member_since', now)
	expected = 'Member Since: {0}'.format(time.ctime(now))

	assert result == expected

def test_format_accounts_works(fixture):
	accounts = [
		dict(account_type='twitter', account_name='batman'),
		dict(account_type='facebook', account_name='Bat Man')
	]
	result = fixture.user.format_accounts(accounts)
	expected = [
		'Share Accounts:',
		fixture.template.format('Twitter: @batman'),
		fixture.template.format('Facebook: Bat Man')
	]

	assert result == expected

def test_lineify_works_without_lists(fixture):
	result = fixture.user.lineify(fixture.selected, False)
	expected = [fixture.user.format(k, v) for k, v in fixture.selected.items()]	

	assert result == expected

def test_lineify_works_with_lists(fixture):
	data = fixture.selected.copy()
	data['full_name'] = ['a', 'b', 'c']
	result = fixture.user.lineify(data, False)
	expected = []
	for key, value in data.items():
		if isinstance(value, list):
			expected.append(fixture.user.format(key))
			expected += [fixture.template.format(i) for i in value]
		else:
			expected.append(fixture.user.format(key, value))

	assert result == expected

def test_lineify_hides_empty_if_wanted(fixture):
	result = fixture.user.lineify({'login': None}, True)

	# better than isinstance(..., list) + 'assert result'
	assert result == []


def test_get_history_works(fixture):
	expected = [fixture.template.format(i) for i in fixture.history]
	result = fixture.user.get_history()
	result.sort()

	assert result == expected


def test_fetch_works_without_history(fixture):
	result = fixture.user.fetch(fixture.only,
								fixture.hide,
								True,
								False,
								False)
	result.sort()

	assert result == fixture.formatted


def test_fetch_works_with_history(fixture):
	result = fixture.user.fetch(fixture.only,
								fixture.hide,
								True,
								True,
								False)
	result[0].sort()
	result[1].sort()
	history = [fixture.template.format(i) for i in fixture.history]
	history.sort()
	expected = [fixture.formatted, history]

	assert result == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import requests

import tests.paths

import lnk.errors
import lnk.tinyurl.command

@pytest.fixture(scope='module')
def fixture():
	return lnk.tinyurl.command.Command('link')

def request(url='http://python.org', response_format='json', version=1):
	return requests.get('http://tiny-url.info/api/v{0}/create'.format(version),
						params=dict(apikey='0BFA4A7B5BDD5BE7780C',
									 format=response_format,
									 provider='tinyurl_com',
									 url=url))

def test_verify_works_for_healthy_response(fixture):
	response = request()
	data = fixture.verify(response, 'even')

	assert data == response.json()

def test_verify_throws_for_http_error(fixture):
	response = request(version=123)
	with pytest.raises(lnk.errors.HTTPError):
		fixture.verify(response, 'even')

def test_verify_throws_for_api_error(fixture):
	response = request('telegram')
	with pytest.raises(lnk.errors.APIError):
		fixture.verify(response, 'even')




#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Data beautification."""

from __future__ import unicode_literals
from __future__ import division

import math
import os
import re
import sys

from collections import namedtuple

import lnk.errors

MAX_WIDTH = 60 # 3/4 of 80
# If we are connected to a terminal right now
if sys.stdin.isatty():
	with os.popen('stty size', 'r') as process:
		output = process.read()
	# Doesn't always work?
	if output:
		MAX_WIDTH = 3 * int(output.split()[1])//4

Line = namedtuple('Line', ['raw', 'escaped'])

def boxify(results):
	"""
	Formats results of command into a box.

	Arguments:
		results (list): A list of results, where each result is a list of lines
						(e.g. all the lines for the statistics of one URL).

	Returns:
		The results in a pretty box, as a string.
	"""
	if not results:
		raise lnk.errors.InternalError('Cannot boxify empty results!')
	results, width = get_escaped(results)
	border = '─' * (width + 2)
	lines = ['┌{0}┐'.format(border)]
	for n, result in enumerate(results):
		m = 0
		# We might add lines into the list mid-iteration
		while m < len(result):
			line = result[m]
			if len(line.escaped) > width:
				result = result[:m] + wrap(line, width) + result[m + 1:]
			else:
				adjusted = ljust(line, width)
				lines.append('│ {0} │'.format(adjusted))
				m += 1
		if n + 1 < len(results):
			lines.append('├{0}┤'.format(border))

	lines += ['└{0}┘'.format(border)]

	return '\n'.join(lines)

def wrap(line, width, indent=None):
	"""
	Wraps a line to a certain width.

	Why not textwrap.wrap? Because in this special situation it is necessary
	to wrap not only the escaped, plain-text string but *at the same time*
	also the raw string containing escape codes, which, however, should not
	count towards the length of a string. This method essentially wraps the
	plain-text string, while at the same time moving also through the raw
	string, but skipping escape-character-sequences when counting characters.

	Arguments:
		line (beauty.Line): The beauty.Line object to wrap.
		width (int): The width to which to wrap.
		indent (str): An indent string with which to indent all lines after the
					  first wrapped one.
	"""
	indent = indent or ' ' * int(math.ceil(width/10))
	escape_code = re.compile(r'\033\[(\d;?)+m')
	r_start = 0
	r = escape_code.match(line.raw)
	r = r.end() if r else 0
	e = 0
	e_start = 0
	wrapped = []
	while e <= len(line.escaped):
		w = width - len(indent) if len(wrapped) else width
		if (e > 0 and (e - e_start) == w) or (e == len(line.escaped)):
			stop = e_start - 1 if e_start else None
			reverse = line.escaped[e - 1:stop:-1]
			boundary = re.search(r'\w(?=\b)', reverse)
			index = e_start + (len(reverse) - boundary.end())
			# Only go backwards to the last word boundary if that
			# is not where we started (else just split it mid-word here).
			# If we went back to the start, we'd get an endless loop
			if e < len(line.escaped) and index != e_start:
				# Go backwards to the last word boundary
				while e > index:
					match = re.match(r'm(;?\d)+\[\033', line.raw[e::-1])
					if match:
						r -= match.end()
					else:
						e -= 1
					r -= 1
			# If we already have one string wrapped
			if len(wrapped) > 0:
				raw = indent + line.raw[r_start:r]
				escaped = indent + line.escaped[e_start:e]
			else:
				raw = line.raw[r_start:r]
				escaped = line.escaped[e_start:e]
			wrapped.append(Line(raw, escaped))
			e_start = e
			r_start = r
		e += 1
		r += 1
		# Skip escape codes for the raw string
		match = escape_code.match(line.raw, r)
		if match:
			r = match.end()

	return wrapped

def ljust(line, width, padding=' '):
	"""
	Adds space-padding to the right of a line.

	ljust won't do because of the escaped/raw thing.

	Arguments:
		line (str): The line to adjust.
		width (int): The minimum width the string must have after this function.
		padding (str): Optionally, the string with which to pad (usually ' ').

	Returns:
		The adjusted raw line.
	"""
	return line.raw + padding * (width - len(line.escaped))

def get_escaped(results):
	"""
	Gets escaped lines for the results passed to boxify().

	Arguments:
		results (list): The results passed to boxify().

	Returns:
		A list of beauty.Line objects with a raw and escaped component.
	"""
	width = 0
	escaped = []
	for result in results:
		lines = []
		for line in result:
			line = escape(line)
			if len(line.escaped) > width:
				width = len(line.escaped)
			lines.append(line)
		escaped.append(lines)

	return escaped, min([MAX_WIDTH, width])

def escape(line):
	"""
	Escapes a raw string.

	Arguments:
		line (str): The raw string to escape.

	Returns:
		A beauty.Line object with the original raw component and the same
		string, but escaped, i.e. without the formatting codes applied by
		ecstasy -- only the text.
	"""
	pattern = re.compile(r'^(.*)'					# anything
						 r'(?:\033\[(?:\d;?)+m)'   	# escape codes
						 r'(.+)'				 	# formatted string
						 r'(?:\033\[(?:\d;?)+m)'   	# escape codes
						 r'(.*)$')					# anything
	line = line.rstrip()
	if '\033' in line:
		match = pattern.search(line)
		if not match:
			raise lnk.errors.InternalError("Could not parse '{0}'".format(line))
		escaped = ''.join([i for i in match.groups() if i])
	else:
		escaped = line

	return Line(line, escaped)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ecstasy
import oauth2client.file
import pyperclip
import pytest
import requests
import threading

try:
	from Queue import Queue
except ImportError:
	from queue import Queue

from collections import namedtuple

import tests.paths
import lnk.googl.link

VERSION = 1
KEY = 'AIzaSyAoXKM_AMBafkXqmVeqJ82o9B9NPCTvXxc'
API = 'https://www.googleapis.com/urlshortener'

LOCK = threading.Lock()
QUEUE = Queue()

def shorten(url='http://python.org'):
	headers = {'content-type': 'application/json'}
	data = '{{"longUrl": "{0}"}}'.format(url)
	params = dict(key=KEY)
	response = requests.post('{0}/v{1}/url'.format(API, VERSION),
							 headers=headers,
							 data=data,
							 params=params)
	return response.json()['id']

def shorten_fmt(destination):
	url = QUEUE.get()
	short = shorten(url)
	formatted = '{0} => {1}'.format(url, short)

	LOCK.acquire()
	destination.append(formatted)
	LOCK.release()


def expand(url='http://goo.gl/Euc5', version=VERSION):
	response = requests.get('{0}/v{1}/url'.format(API, version),
						params=dict(shortUrl=url, key=KEY))

	return response.json()['longUrl']


def expand_fmt(destination):
	url = QUEUE.get()
	expanded = expand(url)
	formatted = '{0} => {1}'.format(url, expanded)

	LOCK.acquire()
	destination.append(formatted)
	LOCK.release()


@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', [
		'link',
		'long',
		'short',
		'bold',
		'long_to_short',
		'short_to_long'
		])

	link = lnk.googl.link.Link(raw=True)
	link.credentials = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	url = 'https://www.github.com/goldsborough/lnk'
	short = shorten(url)
	bold = ecstasy.beautify('<{0}>'.format(short), ecstasy.Style.Bold)
	long_to_short = '{0} => {1}'.format(url, short)
	short_to_long = '{0} => {1}'.format(short, url)

	return Fixture(link, url, short, bold, long_to_short, short_to_long)


def test_copy_copies_to_clipboard_if_copy_true(fixture):
	fixture.link.copy(True, fixture.short)

	assert pyperclip.paste() == fixture.short


def test_copy_copies_only_first_url(fixture):
	assert fixture.link.already_copied

	fixture.link.copy(True, 'a')
	fixture.link.copy(True, 'b')
	fixture.link.copy(True, 'c')

	assert pyperclip.paste() == fixture.short


def test_copy_copies_to_clipboard_if_copy_false(fixture):
	pyperclip.copy('original')
	fixture.link.copy(False, fixture.short)

	assert pyperclip.paste() == 'original'


def test_copy_makes_copied_url_bold(fixture):
	fixture.link.already_copied = False
	returned_url = fixture.link.copy(True, fixture.short)

	assert returned_url == fixture.bold


def test_get_long_expands_well(fixture):
	result = fixture.link.get_long(fixture.short)

	assert result == fixture.long


def test_get_short_shortens_well(fixture):
	short = fixture.link.get_short(fixture.long)
	result = fixture.link.get_long(short)

	assert result == fixture.long


def test_shorten_formats_well(fixture):
	result = []
	fixture.link.queue.put(fixture.long)
	fixture.link.shorten(result, False)
	result = result[0].split()

	assert result[0] == fixture.long
	assert result[1] == '=>'
	assert result[2].startswith('https://goo.gl/')


def test_expand_formats_well(fixture):
	result = []
	fixture.link.queue.put(fixture.short)
	fixture.link.expand(result, False)

	assert result[0] == fixture.short_to_long


def test_shorten_urls_works_for_single_url(fixture):
	result = fixture.link.shorten_urls(False, True, [fixture.long])
	result = result[0].split()

	assert result[0] == fixture.long
	assert result[1] == '=>'

	expanded = fixture.link.get_long(result[2])

	assert expanded == result[0]


def test_shorten_urls_works_for_many_urls(fixture):
	urls = [
		'http://facebook.com/',
		'http://google.com/',
		'http://python.org/'
	]
	result = fixture.link.shorten_urls(False, True, urls)
	expected = []
	threads = []
	for url in urls:
		QUEUE.put(url)
		thread = threading.Thread(target=shorten_fmt, args=(expected,))
		thread.daemon = True
		threads.append(thread)
		thread.start()

	for thread in threads:
		thread.join(timeout=10)

	print(sorted(result), sorted(expected))

	for got, wanted in zip(sorted(result), sorted(expected)):
		got = got.split()
		wanted = wanted.split()

		assert got[0] == wanted[0]
		assert got[1] == wanted[1] == '=>'

		expanded = fixture.link.get_long(got[2])
		assert expanded == got[0] == wanted[0]


def test_expand_urls_works_for_single_url(fixture):
	result = fixture.link.expand_urls(False, [fixture.short])

	assert result[0] == fixture.short_to_long


def test_expand_urls_works_for_many_urls(fixture):
	urls = [
		'http://goo.gl/Up0wrT',
		'http://goo.gl/4Nuepy',
		'http://goo.gl/bYm2EP'
	]
	result = fixture.link.expand_urls(False, urls)
	expected = []
	threads = []
	for url in urls:
		QUEUE.put(url)
		thread = threading.Thread(target=expand_fmt, args=(expected,))
		thread.daemon = True
		thread.start()
		threads.append(thread)

	for thread in threads:
		thread.join(timeout=10)

	print(result, expected)

	assert sorted(result) == sorted(expected)


def test_shorten_urls_warns_about_url_without_protocol(fixture, capsys):
	fixture.link.shorten_urls(False, False, ['google.com'])
	out = capsys.readouterr()

	assert out
	assert out[0].startswith("\aWarning: Prepending 'http://' to")


def test_fetch_works(fixture):
	result = fixture.link.fetch(False,
								True,
								[fixture.short],
								[fixture.long],
								False)
	expected = [fixture.long_to_short, fixture.short_to_long]
	for got, wanted in zip(sorted(result), sorted(expected)):
		got = got.split()
		wanted = wanted.split()

		assert got[0] == wanted[0]
		assert got[1] == wanted[1] == '=>'

		if got[0] == fixture.long:
			expanded = fixture.link.get_long(got[2])
			assert expanded == got[0] == wanted[0]


def test_fetch_correct_output_if_raw_false_pretty_false(fixture):
	fixture.link.raw = False
	result = fixture.link.fetch(False,
								True,
								[fixture.short],
								[fixture.long],
								False)
	expected = '\n'.join([fixture.short_to_long, fixture.long_to_short])

	return result == expected



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

import os.path

from .manager import Manager, get

CONFIG_PATH = os.path.join(
	os.path.abspath(
		os.path.dirname(
			os.path.dirname(
				os.path.dirname(
					os.path.abspath(__file__)
				)
			)
		)
	),
	'config'
)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""The command-line interface to the config command."""

import click

import lnk.config.manager
import lnk.config.configure
import lnk.errors

allowed_wich = lnk.config.manager.get('lnk', 'services') + ['lnk', 'config']

@click.command()
@click.argument('which',
				default='lnk',
				type=click.Choice(allowed_wich))
@click.argument('command',
				default=None,
				required=False)
@click.option('-k',
			  '--key',
			  nargs=1,
			  multiple=True,
			  metavar='KEY',
			  help='A key to show or manipulate.')
@click.option('-v',
			  '--value',
			  nargs=1,
			  multiple=True,
			  metavar='VALUE',
			  help='A new value for a key.')
@click.option('-q/-l',
			  '--quiet/--loud',
			  default=False,
			  help='Whether to visualize changes.')
@click.option('-a',
			  '--all',
			  '--all-keys',
			  is_flag=True,
			  help='Whether to show all keys.')
def main(which, command, key, value, quiet, all_keys):
	"""Configuration interface."""
	if command:
		if which == 'lnk':
			what = 'lnk has no commands to configure.'
			hint = "If '{0}' is a key, you ".format(command)
			hint += "probably wanted to use '-k {0}'.".format(command)
			raise lnk.errors.UsageError(what, Hint=hint)
		elif command not in lnk.config.get(which, 'commands').keys():
			what = "No such command '{0}' for {1}.".format(command, which)
			raise lnk.errors.UsageError(what)
	if key or all_keys:
		lnk.errors.catch(lnk.config.configure.echo,
						 which,
						 command,
						 key,
						 value,
						 quiet,
						 all_keys)
	else:
		main(['--help'])



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Contains the base-class for all bit.ly commands."""

import lnk.config
import lnk.errors

from lnk.abstract import AbstractCommand

class Command(AbstractCommand):
	"""
	Base-class for all bit.ly commands.

	Configures the AbstractCommand base class for all commands in the
	entire application, which needs information about the service being
	used. Moreover sets up the necessary parameters needed for any request
	to the bit.ly API (the OAuth2 access token). 

	Attributes:
		parameters (dict): The necessary parameters for any request to the
						   bit.ly API.
	"""
	def __init__(self, which):
		"""
		Raises:
			errors.AuthorizationError: If the OAuth2 access token cannot be
									   retrieved from the configuration file.
		"""
		super(Command, self).__init__('bitly', which)
		with lnk.config.Manager('bitly') as manager:
			if not manager['key'] and which != 'key':
				raise lnk.errors.AuthorizationError('bitly')
			self.parameters = {'access_token': manager['key']}

	@staticmethod
	def verify(response, what, inner=None):
		"""
		Verifies an HTTP-response from the bit.ly API.

		Overrides the 'pure-virtual' (i.e. not-implemented) base method
		from AbstractCommand. If the verification finds no faults in the
		response, the data is returned.

		Arguments:
			response (requests.Response): The HTTP response to a request
										  to the bit.ly API.
			what (str): A human-readable string representing what the request
						was for, such that if there is an error in the response,
						an errors.HTTPError or errors.APIError is raised with
						the message 'Could not <what>.'
			inner (str): If it is expected that data has an inner layer, the
						 key of that inner layer to retrieve the data directly.

		Returns:
			The actual data of the response, if no fault was found.

		Raises:
			errors.HTTPError: If it was found that there was an HTTP-related
							  exception, such as a faulty URL or other badness.
			errors.APIError: If it was found that there was an exception
							 related to the API itself.
		"""
		if not str(response.status_code).startswith('2'):
			raise lnk.errors.HTTPError('Could not {0}.'.format(what),
								   response.status_code,
								   response.reason)
		response = response.json()
		if not str(response['status_code']).startswith('2'):
			raise lnk.errors.HTTPError('Could not {0}.'.format(what),
								   response['status_code'],
								   response['status_txt'])
		data = response['data']
		if inner:
			data = data[inner][0]
		if 'error' in data:
			what = 'Could not {0}!'.format(what)
			raise lnk.errors.APIError(what, data['error'])

		return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pytest
import requests

import tests.paths

import lnk.bitly.command
import lnk.config
import lnk.errors

VERSION = 3
API = 'https://api-ssl.bitly.com'
with open(os.path.join(tests.paths.TEST_PATH, 'bitly', 'token')) as source:
	ACCESS_TOKEN = source.read()


@pytest.fixture(scope='module')
def fixture():
	return lnk.bitly.command.Command('link')


def shorten(url='http://python.org', version=VERSION):
	return requests.get('{0}/v{1}/shorten'.format(API, version),
						params=dict(access_token=ACCESS_TOKEN,
									longUrl=url))


def expand(url):
	return requests.get('{0}/v{1}/expand'.format(API, VERSION),
						params=dict(access_token=ACCESS_TOKEN,
									shortUrl=url))


def test_throws_when_not_yet_authenticated():
	with lnk.config.Manager('bitly', write=True) as manager:
		old = manager['key']
		manager['key'] = None
	with pytest.raises(lnk.errors.AuthorizationError):
		lnk.bitly.command.Command('link')
	with lnk.config.Manager('bitly', write=True) as manager:
		manager['key'] = old


def test_initializes_well(fixture):
	assert hasattr(fixture, 'parameters')


def test_verify_works_for_healthy_response(fixture):
	response = shorten()
	result = fixture.verify(response, 'even')

	assert result == response.json()['data']


def test_verify_throws_for_http_error(fixture):
	response = shorten(version=123)
	with pytest.raises(lnk.errors.HTTPError):
		fixture.verify(response, 'even')

	response = shorten('invalid_uri')
	with pytest.raises(lnk.errors.HTTPError):
		fixture.verify(response, 'even')


def test_verify_throws_for_api_error(fixture):
	response = expand('google.com')

	with pytest.raises(lnk.errors.APIError):
		fixture.verify(response, 'even', 'expand')



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Link shortening and expansion for the tinyurl client."""

import click
import ecstasy
import pyperclip
import re

import lnk.beauty
import lnk.errors

from lnk.tinyurl.command import Command

def echo(*args):
	"""
	Executes a link command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to an
						 Link instance's fetch() method.
	"""
	click.echo(Link().fetch(*args))

class Link(Command):
	"""
	Class to shorten or expand a url using tinyurl.

	This class can shorten a long url to a shortened tinyurl url.

	Attributes:
		raw (bool): Whether to prettify the output or
					return it raw, for internal use.
		already_copied (bool): Flag set the first time a url is copied
							   to the clipboard, such that only the first
							   is copied.
		http (regex): A compiled regular-expression object matching a
					  HTTP(S) protocol, for URL-checking.
	"""
	def __init__(self, raw=False):
		super(Link, self).__init__('link')
		self.raw = raw
		self.already_copied = False
		self.http = re.compile(r'https?://')

	def fetch(self, copy, quiet, urls, pretty):
		"""
		Fetches a short or expanded link.

		Arguments:
			copy (bool): Whether or not to copy the first shortened or
						 expanded link to the clipboard.
			quiet (bool): Whether or not to swallow warnings about urls
						  having to be modified because they do not have
						  a protocol specified (i.e. no 'http://').
			urls (tuple): A tuple of long urls to shorten.
			pretty (bool): Whether or not to make the output pretty (in a box).

		Returns:
			A raw list of lines for other internal use if self.raw is True,
			else a pretty list in a box if pretty is True, else the same
			list as in the first case, but joined to a string for output.
		"""
		result = []
		threads = []
		for url in urls:
			t = self.new_thread(self.shorten, result, copy, quiet, pretty, url)
			threads.append(t)
		self.join(threads)

		if self.raw:
			return result
		return lnk.beauty.boxify([result]) if pretty else '\n'.join(result)

	def shorten(self, result, copy, quiet, pretty, url):
		"""
		Shortens a long url.

		Arguments:
			result (list): The list to which to append the shortened url.
			copy (bool): Whether or not to copy the first url to the clipboard.
			quiet (bool): Whether or not to swallow warnings about urls
						  having to be modified because they do not have
						  a protocol specified (i.e. no 'http://').
			pretty (bool): Whether or not to make the url pretty. If yes,
						   the formatted string appended to the result list
						   follows the schema '<long> => <url>'.
			urls (tuple): The tuple of long urls to shorten.
		"""
		if not self.http.match(url):
			url = 'http://{0}'.format(url)
			if not quiet:
				lnk.errors.warn("Prepending 'http://' to '{0}'".format(url))
		short = self.request(url)
		formatted = self.copy(copy, short)
		if pretty:
			formatted = '{0} => {1}'.format(url, formatted)
		with self.lock:
			result.append(formatted)

	def request(self, url):
		"""
		Requests and returns a short url for a long one.

		Arguments:
			url (str): The long url to shorten.

		Returns:
			The shortened link.
		"""
		response = self.get(self.endpoints['create'], dict(url=url))
		response = self.verify(response, "shorten url '{0}'".format(url))

		return response['shorturl']

	def copy(self, copy, url):
		"""
		Copies a url to the clipboard if possible.

		"If possible" means the url is only copied if no other url
		has so-far been copied to the clipboard during the same
		fetch() call, and also only if the 'copy' is set. If the url
		is copied, it is formatted using ecstasy to make it bold.

		Arguments:
			copy (bool): Whether or not to allow copying
						 to the clipboard at all.
			url (str): The url to (possibly) copy.

		Returns:
			The original url if the 'copy' flag is unset or another link
			has already been copied to the clipboard during the same call
			to fetch(), else the url, formatted with ecstasy to appear bold
			in the terminal.
		"""
		if copy and not self.already_copied:
			self.already_copied = True
			pyperclip.copy(url)
			url = ecstasy.beautify('<{0}>'.format(url), ecstasy.Style.Bold)

		return url



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Handles the oauth2 authorization procedure for the goo.gl API."""

import click
import ecstasy
import oauth2client.file
import oauth2client.tools
import os.path
import webbrowser

from oauth2client.client import OAuth2WebServerFlow

import lnk.errors
import lnk.config

from lnk.googl.command import Command

CLIENT_ID = '324510822959-qcp2d36tkl07v2dfqde3ungmpii4qv96.apps.googleusercontent.com'
CLIENT_SECRET = 'r7-N-xmGbfePxZP3q6QEXu1Z'
SCOPE = 'https://www.googleapis.com/auth/urlshortener'
REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'

def echo(*args):
	"""
	Executes a key command.

	Arguments:
		args (variadic): The arguments to pass to a
						 Key instance's fetch() method.
	"""
	Key().fetch(*args)

class Key(Command):
	"""
	Handles oauth2 authorization for the goo.gl API.

	The authorization procedure involves exchaning a temporary authorization
	code for an oauth2 token. When executing this command, the user will be
	redirected to a web-page where he or she can authorize lnk to access his
	or her private goo.gl information. Once done the user gets an
	authorization-code which he or she must then paste into the command-line,
	where it is fetched by lnk and stored for authorization during HTTP requests.

	Attributes:
		raw (bool): Whether or not to return the raw output for internal use
					or the prettified output for outside representation.
		flow (oauth2client.client.OAuth2WebServerFlow): An object from Google's
										API-library that is used for the entire
										authorization-code-to-oauth2-token
										exchange procedure.
		credentials_path (str): The path to the credentials file in the config
								folder.
	"""
	def __init__(self, raw=False):
		super(Key, self).__init__('key')
		self.raw = raw
		self.flow = OAuth2WebServerFlow(client_id=CLIENT_ID,
										client_secret=CLIENT_SECRET,
										scope=SCOPE,
										redirect_uri=REDIRECT_URI)
		self.credentials_path = os.path.join(lnk.config.CONFIG_PATH, '.credentials')


	def fetch(self, _):

		# First request an authorization token as part of the oauth handshake.
		authorize_url = self.flow.step1_get_authorize_url()

		# The user is redirected to Google's authorization page where permission
		# must be granted to the lnk application to access the user's data
		click.echo("Redirecting you to Google's authorization page ...")
		webbrowser.open(authorize_url)

		code = click.prompt('Please enter the authorization code',
							hide_input=True)

		# Now exchange the authorization token for credentials, i.e.
		# an access_token that must be passed as part of the oauth protocol
		# with any API call, and a refresh_token that can be used to refresh
		# the access_token, which expires after a certain time (~ 1 hour).
		credentials = self.flow.step2_exchange(code)

		storage = oauth2client.file.Storage(self.credentials_path)
		storage.put(credentials)

		success = ecstasy.beautify('<Success!>', ecstasy.Color.Magenta)
		click.echo(success)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""setup.py script for setuptools."""

import json
import os.path
import re

import setuptools
import setuptools.command.install

with open(os.path.join('lnk', '__init__.py')) as init:
	INIT = init.read()


def get(thing):
	"""Retrieves a setup value."""
	if thing == 'readme':
		with open('README.rst') as readme:
			return readme.read()
	pattern = r'^__{0}__\s*=\s*[\'"]([^\'"]*)[\'"]'.format(thing)
	match = re.search(pattern, INIT, re.MULTILINE)

	return match.group(1)


class Install(setuptools.command.install.install):
	"""Subclass of setuptool's install class to have a post-install task."""

	def run(self):
		setuptools.command.install.install.run(self)
		self.execute(self.reset, [], msg='Running post-install task...')

	@staticmethod
	def reset():
		"""Resets any old authorization keys for a new user."""
		for_bitly = os.path.join('config', 'bitly.json')
		with open(for_bitly) as source:
			config = json.load(source)
		config['key'] = None
		with open(for_bitly, 'w') as destination:
			json.dump(config, destination, indent=4)
		for_googl = os.path.join('config', '.credentials')
		if os.path.exists(for_googl):
			os.remove(for_googl)


setuptools.setup(
	cmdclass=dict(install=Install),
	name=get('title'),
	version=get('version'),
	description='A command-line URL-shortening client.',
	long_description=get('readme'),
	author=get('author'),
	author_email=get('email'),
	url=get('url'),
	license=get('license'),
	classifiers=[
		'Development Status :: 4 - Beta',

		'Intended Audience :: Developers',
		'Intended Audience :: End Users/Desktop',

		'Topic :: Internet',
		'Topic :: Desktop Environment',

		'License :: OSI Approved :: MIT License',

		'Programming Language :: Python',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.2',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
	],
	keywords='lnk, url-shortening, bitly, googl, tinyurl',
	include_package_data=True,
	package_data=dict(lnk=[
		'../README.rst',
		'../LICENSE',
		'../Makefile',
		'../config/*',
		'../docs/source/*.rst',
		'../docs/source/conf.py',
		'../docs/Makefile',
		]),
	packages=setuptools.find_packages(exclude=['scripts']),
	install_requires=[
		'click==4.1',
		'coverage==3.7.1',
		'ecstasy==0.1.3',
		'google-api-python-client==1.4.2',
		'pyperclip==1.5.11',
		'requests==2.7.0',
		'ordereddict==1.1',
		'enum34==1.0.4'
	],
	test_suite='tests',
	tests_require=[
		'pytest==2.7.2',
		'pytest-cache==1.0',
		'python-coveralls==2.5.0',
		'tox==2.1.1'
	],
	entry_points=dict(console_scripts=['lnk = lnk.cli:main'])
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime
import oauth2client.file
import pytest
import requests

from collections import namedtuple

import tests.paths
import lnk.googl.info

VERSION = 1
KEY = 'AIzaSyAoXKM_AMBafkXqmVeqJ82o9B9NPCTvXxc'
API = 'https://www.googleapis.com/urlshortener'

def request(url):
	response = requests.get('{0}/v{1}/url'.format(API, VERSION),
							params=dict(projection='FULL',
										shortUrl=url,
										key=KEY))
	return response.json()


@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', [
		'info',
		'only',
		'hide',
		'sets',
		'urls',
		'expanded',
		'data',
		'formatted'
		])
	urls = ['http://goo.gl/9RJxHk', 'https://goo.gl/IpUmJn']
	expanded = ['http://python.org/', 'http://github.com/']
	info = lnk.googl.info.Info(raw=True)
	info.credentials = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	only = ['created', 'expanded']
	hide = ['status']
	sets = {
		"created": "created", 
		"expanded": "longUrl"
	}

	data = request(urls[0])
	selected = {}
	# No dictionary comprehension in Python 2.6
	for key, value in data.items():
		if key in sets.values():
			selected[key] = value
	formatted = ['URL: {0}'.format(urls[0])]
	for key, value in selected.items():
		formatted.append(info.format(key, value))

	return Fixture(info,
					only,
					hide,
					sets,
					urls,
					expanded,
					selected,
					formatted)

def test_initializes_well(fixture):

	assert hasattr(fixture.info, 'sets')
	assert hasattr(fixture.info, 'reverse')

	reverse = dict((v, k) for k, v in fixture.info.sets.items())

	assert fixture.info.reverse == reverse
	assert all(i in fixture.info.sets for i in fixture.only)

def test_requests_well(fixture):
	expected = request(fixture.urls[0])
	result = fixture.info.request(fixture.urls[0])

	assert result == expected

def test_format_sets_keys_well(fixture):
	result = fixture.info.format('status', 'OK')
	expected = 'Status: OK'

	assert result == expected

def test_format_parses_times_well(fixture):
	now = datetime.datetime.now()
	result = fixture.info.format(fixture.sets['created'],
								 now.isoformat())
	expected = 'Created: {0}'.format(now.ctime())

	assert result == expected

def test_lineify_works_without_lists(fixture):
	result = fixture.info.lineify(fixture.urls[0], fixture.data)

	assert result == fixture.formatted

def test_gets_info_well(fixture):
	fixture.info.queue.put(fixture.urls[0])
	result = []
	fixture.info.get_info(fixture.sets.values(), result)

	assert result[0] == fixture.formatted

def test_fetches_well(fixture):
	result = fixture.info.fetch(fixture.only,
								fixture.hide,
								fixture.urls)

	data = request(fixture.urls[1])
	selection = {}
	# No dictionary comprehension in Python 2.6
	for key, value in data.items():
		if key in fixture.sets.values():
			selection[key] = value
	second = ['URL: {0}'.format(fixture.urls[1])]
	second += [fixture.info.format(k, v) for k, v in selection.items()]
	expected = [fixture.formatted, second]

	result[0].sort()
	result[1].sort()
	expected[0].sort()
	expected[1].sort()

	assert sorted(result) == sorted(expected)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""The main entry-point to lnk from the command-line."""

import click
import os
import sys

import lnk.config
import lnk.errors

class Main(click.MultiCommand):
	"""
	The main command-group of the lnk application.

	This class exists because the necessary subcommands of lnk
	(i.e. the commands for the services, such as bitly or googl)
	live in other files and click does not handle that well. There
	is no way to register commands from other files, so this class,
	which derives from click.MultiCommand, must dynamically load
	and compile the files in which the other commands live and
	extract the necessary functions so that click can execute each
	subcommand as if it lived in this file. It also handles argument
	forwarding and some string-handling (e..g so that googl and goo.gl both
	mean the same thing from the CLI). Its all a bit of a hack.

	Attributes:
		commands (dict): A mapping between the names of commands and their
						 respective functions.
		default (str): The name of the default service (e.g. 'bitly').
	"""
	def __init__(self):
		super(Main, self).__init__(context_settings=dict(
								  ignore_unknown_options=True))
		self.commands = {}
		with lnk.config.Manager('lnk') as manager:
			self.default = manager['settings']['service']
			for command in manager['services'] + ['config']:
				command = command.replace('.', '') # goo.gl -> googl
				self.commands[command] = self.get_function(command)

	def format_usage(self, context, *args):
		# Hack to make the main script's name
		# appear as 'lnk' rather than 'main.py'
		context.info_name = 'lnk'

		return super(Main, self).format_usage(context, *args)

	def invoke(self, context):
		"""
		Invokes a command.

		This method does some argument pre-processing, such as inserting
		the name of the default command if no other command was specified.
		It also handles the problem of dots in the names of commands (e.g.
		goo.gl and gool).
		"""
		escaped = context.args[0].replace('.', '')
		if escaped not in self.commands.keys():
			context.args.insert(0, self.default.replace('.', ''))
		else:
			context.args[0] = escaped
		super(Main, self).invoke(context)

	def list_commands(self, context):
		"""Returns the names of all available subcommands."""
		return self.commands.keys()

	def get_command(self, context, name):
		"""Returns the function for a given subcommand-name."""
		if name not in self.commands:
			clue = lnk.errors.Message('Did you mess up the default settings?',
									  level=0)
			try_message = lnk.errors.Message("See what 'lnk config -k service'"
									  		 " says.", level=1)
			raise lnk.errors.UsageError('Invalid default service.',
										Clue=clue,
										Try=try_message)
		return self.commands[name]

	@staticmethod
	def get_function(service):
		"""
		Returns the 'main' command-function for a given service.

		Arguments:
			service (str): The name of the service (e.g. tinyurl).

		Returns:
			The main() function of that service (retrieved from
			lnk.<service>.cli).
		"""
		namespace = {}
		directory = os.path.dirname(__file__)
		filename = os.path.join(directory, service, 'cli.py')
		with open(filename) as source:
			code = compile(source.read(), filename, 'exec')
			eval(code, namespace, namespace)

		return namespace['main']

def get_verbosity(count, level):
	"""
	Returns the appropriate verbosity level.

	Defined here for use by all cli modules (e.g. bitly.cli).

	Arguments:
		count (int): The verbosity count (from the -vvv flags).
		level (int): The verbosity level (from the --level flag).

	Returns:
		The default verbosity if both count and level are zero, else
		the level and only if the level is also zero, the count.
	"""
	if count == 0 and not level:
		return lnk.config.get('lnk', 'settings')['verbosity']
	elif level:
		return level
	return count

def main():
	"""
	Insantiates a Main object and executes it.

	The 'standalone_mode' is so that click doesn't handle usage-exceptions
	itself, but bubbles them up.

	Arguments:
		args (tuple): The command-line arguments.
	"""
	args = sys.argv[1:]
	# If stdin is not empty (being piped to)
	if not sys.stdin.isatty():
		args += sys.stdin.readlines()
	command = Main()
	catch = lnk.errors.Catch(1)
	catch.catch(command.main, args, standalone_mode=False)

if __name__ == '__main__':
	main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy
import ecstasy
import os
import pytest
import requests
import time

from collections import namedtuple

import tests.paths
import lnk.bitly.info

VERSION = 3
API = 'https://api-ssl.bitly.com/v{0}'.format(VERSION)
with open(os.path.join(tests.paths.TEST_PATH, 'bitly', 'token')) as source:
	ACCESS_TOKEN = source.read()

def request_info(url):
	response = requests.get('{0}/info'.format(API),
							params=dict(shortUrl=url,
										access_token=ACCESS_TOKEN))

	return response.json()['data']['info'][0]

def request_history(url):
	response = requests.get('{0}/user/link_history'.format(API),
							params=dict(access_token=ACCESS_TOKEN,
										link=url))

	return response.json()['data']['link_history'][0]

@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', [
		'info',
		'only',
		'hide',
		'sets',
		'urls',
		'expanded',
		'data',
		'formatted',
		'template'
		])

	urls = ['http://bit.ly/1V0HZR7', 'http://bit.ly/1V0HPcB']
	expanded = ['http://google.com/', 'http://stackoverflow.com/']
	info = lnk.bitly.info.Info(raw=True)
	only = ['created', 'expanded', 'user', 'tags']
	hide = ['modified']
	sets = {
		'created': 'created_at', 
        'expanded': 'long_url', 
        'user': 'created_by',
        'tags': 'tags'
	}

	data = request_info(urls[0])
	data.update(request_history(urls[0]))
	selection = {}
	# No dictionary comprehension in Python 2.6
	for key, value in data.items():
		if key in sets.values():
			selection[key] = value
	print(selection)
	formatted = ['URL: {0}'.format(urls[0])]
	template = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	for key, value in selection.items():
		if not value:
			formatted.append(info.format(key, 'None'))
		if isinstance(value, list):
			formatted.append('{0}: '.format(key.title()))
			formatted += [template.format(t) for t in value]
		else:
			formatted.append(info.format(key, value))

	return Fixture(info,
					only,
					hide,
					sets,
					urls,
					expanded,
					selection,
					formatted,
					template)

def test_initializes_well(fixture):

	assert hasattr(fixture.info, 'sets')
	assert hasattr(fixture.info, 'reverse')

	reverse = {}
	for key, value in fixture.info.sets.items():
		reverse[value] = key

	assert fixture.info.reverse == reverse

	assert all(i in fixture.info.sets for i in fixture.only)

def test_requests_info_well(fixture):
	expected = request_info(fixture.urls[0])
	result = fixture.info.request_info(fixture.urls[0])

	assert result == expected

def test_requests_history_well(fixture):
	expected = request_history(fixture.urls[0])
	result = fixture.info.request_history(fixture.urls[0])

	assert result == expected

def test_format_setss_keys_well(fixture):
	result = fixture.info.format(fixture.sets['user'], 'Borat')
	expected = 'User: Borat'

	assert result == expected

def test_format_parses_times_well(fixture):
	now = time.time()
	result = fixture.info.format(fixture.sets['created'], now)
	expected = 'Created: {0}'.format(time.ctime(now))

	assert result == expected

def test_format_handles_privacy_well(fixture):
	result = fixture.info.format('private', 'Yes')
	expected = 'Private: Yes'

	assert result == expected

def test_format_handles_booleans_well(fixture):
	result = fixture.info.format('private', False)
	expected = 'Private: No'

	assert result == expected

def test_format_handles_None_values_well(fixture):
	result = fixture.info.format(fixture.sets['expanded'], None)
	expected = 'Expanded: None'

	assert result == expected

def test_format_does_not_format_value_if_list(fixture):
	result = fixture.info.format('tags', ['a', 'b', 'c'])
	expected = 'Tags: '

	assert result == expected

def test_lineify_works_without_lists(fixture):
	data = copy.deepcopy(fixture.data)
	del data['tags']
	expected = [fixture.formatted[0]]
	expected += [fixture.info.format(k, v) for k, v in data.items()]
	result = fixture.info.lineify(fixture.urls[0], data, False)

	assert result == expected

def test_lineify_works_with_lists(fixture):
	result = fixture.info.lineify(fixture.urls[0],
								  fixture.data,
								  False)

	assert result == fixture.formatted

def test_lineify_hides_empty_if_wanted(fixture):
	result = fixture.info.lineify(fixture.urls[0],
								  {'Thing': None},
								  True)

	assert result == [fixture.formatted[0]]

def test_requests_well(fixture):
	fixture.info.queue.put(fixture.urls[0])
	result = []
	fixture.info.request(fixture.sets.values(), result, False)

	print(result, fixture.formatted)

	assert result[0] == fixture.formatted

def test_fetches_well(fixture):
	result = fixture.info.fetch(fixture.only,
								fixture.hide,
								False,
								fixture.urls)

	data = request_info(fixture.urls[1])
	data.update(request_history(fixture.urls[1]))
	selection = {}
	# No dictionary comprehension in Python 2.6
	for key, value in data.items():
		if key in fixture.sets.values():
			selection[key] = value
	second = ['URL: {0}'.format(fixture.urls[1])]
	for key, value in selection.items():
		second.append(fixture.info.format(key, value))
		if isinstance(value, list):
			second += [fixture.template.format(i) for i in value]
	expected = [fixture.formatted, second]

	result[0].sort()
	result[1].sort()
	expected[0].sort()
	expected[1].sort()

	assert sorted(result) == sorted(expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Handles the oauth2 authorization procedure for the bit.ly API."""

import click
import ecstasy
import simplejson

import lnk.beauty
import lnk.config
import lnk.errors

from lnk.bitly.command import Command

CLIENT_ID = '049118cdcf01c6b267f77582990d48da9127259a'
CLIENT_SECRET = 'b5257382fd1c7118688cae7942d5128b4d995c9e'

def echo(*args):
	"""
	Executes a key command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 Key instance's fetch() method.
	"""
	click.echo(Key().fetch(*args), nl=False)

class Key(Command):
	"""
	Handles oauth2 authorization for the bit.ly API.

	The bulk of the work is done by the user in the command-line interface.
	When 'lnk bitly --key' is run (or 'lnk --key' if bitly is the default),
	the user is prompted for his or her login and password. This information
	is passed to this class, which then fetches an oauth2 access-token using
	the appropriate exchange procedure.

	Attributes:
		raw (bool): Whether or not to return the raw output for internal use
					or the prettified output for outside representation.
	"""
	def __init__(self, raw=False):
		super(Key, self).__init__('key')
		self.raw = raw
		self.parameters['client_id'] = CLIENT_ID
		self.parameters['client_service'] = CLIENT_SECRET

	def fetch(self, _, login, password, show, who):
		"""
		Fetches an oauth2 access-token for a given login and password.

		Arguments:
			_: This is the --generate flag, just a dummy parameter because it
			   must be passed from the command-line to initiate the procedure
			   but has no actual use here.
			login (str): The user's login (username).
			password (str): The user's password.
			show (bool): Whether or not to return the access-token for output
						 once retrieved.

		Returns:
			If the show flag is set, the raw key as a string if self.raw
			is True, else the key in a pretty box. If the show flag is unset,
			an empty string is returned (such that nothing appears in the
			command-line output).
		"""
		if who:
			who = lnk.config.get('bitly', 'login')
			return '{0}\n'.format(who) if who else 'Nobody.\n'

		if not login:
			login = click.prompt('Login')
		if not password:
			password = click.prompt('Password', hide_input=True)

		key = self.request(login, password)
		with lnk.config.Manager('bitly', write=True) as manager:
			manager['key'] = key
			manager['login'] = login

		success = ecstasy.beautify('<Success!>', ecstasy.Color.Magenta)
		click.echo(success)

		if show:
			if self.raw:
				return key
			return lnk.beauty.boxify([[key]]) + '\n'
		return ''

	def request(self, login, password):
		"""
		Requests an oauth2 token from the bit.ly API.

		Arguments:
			login (str): The user's login (username).
			password (str): The user's password.

		Returns:
			The oauth2 token/key, if the request was successful.

		Raises:
			Any errors thrown by the verify() method.
		"""
		response = self.post(self.endpoints['oauth'],
							 authorization=(login, password))

		key = self.verify(response, 'generate an API key')

		return key

	@staticmethod
	def verify(response, what):
		"""
		Verifies the HTTP response following the request for an oauth2 token.

		Arguments:
			response (requests.Response): The HTTP-response from the request
										  for an oauth2 token.
			what (str): A string-representation of what the request was for,
						such that if an error is found and raised, its message
						is 'Could not <what>' (e.g. 'generate an API key').

		Returns:
			The plain, string oauth2 access-token.

		Raises:
			errors.HTTPError: If an HTTP-related fault was found in the response.
		"""
		# If the request succeeded, the response is in
		# text format, so the json decoding would fail,
		# if there was an error we have to retrieve it in json
		try:
			response = response.json()
			if not str(response['status_code']).startswith('2'):
				raise lnk.errors.HTTPError('Could not {0}.'.format(what),
									   response['status_code'],
									   response['status_txt'])
		except simplejson.scanner.JSONDecodeError:
			pass

		return str(response.text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime
import ecstasy
import httplib2
import oauth2client.file
import os.path
import pytest
import requests

from collections import namedtuple

import lnk.errors
import tests.paths
import lnk.googl.history

VERSION = 1
API = 'https://www.googleapis.com/urlshortener'

def timestamp(time_range, base=None):
	delta = {
		"minute": datetime.timedelta(minutes=1), 
		"hour": datetime.timedelta(hours=1), 
		"day": datetime.timedelta(days=1),
		"week": datetime.timedelta(weeks=1), 
		"month": datetime.timedelta(days=30.4368),
		"year": datetime.timedelta(days=365.242199)
	}
	offset = time_range[0] * delta[time_range[1]]
	base = base or datetime.datetime.now()

	return base - offset

def get_token():
	storage = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	credentials = storage.get()
	if credentials.access_token_expired:
		credentials.refresh(httplib2.Http())
	token = credentials.access_token
	storage.put(credentials)

	return token


def request():
	token = get_token()
	url = '{0}/v{1}/url/history'.format(API, VERSION)
	urls = []
	data = {'nextPageToken': None}
	while 'nextPageToken' in data:
		response = requests.get(url,
								params={
								'start-token': data['nextPageToken'],
								'access_token': token
								})
		data = response.json()
		urls += data['items']

	return urls


def process(urls):
	return lnk.googl.history.History.process(urls)


def filter_urls(urls, start=None, end=None):
	if start:
		start = timestamp(start)
	else:
		start = datetime.datetime.min
	if end:
		end = timestamp(end)
	else:
		end = datetime.datetime.now()
	filtered = []
	for url in urls:
		if url.created >= start and url.created <= end:
			filtered.append(url)

	return filtered


@pytest.fixture(scope='module')
def fixture():
	
	Fixture = namedtuple('Fixture', [
		'history',
		'all_urls',
		'last',
		'last_urls',
		'ranges',
		'ranges_urls',
		'template',
		'url',
		'dummies',
		'short_dummies'
		])

	urls = process(request())

	history = lnk.googl.history.History(raw=True)

	history.credentials = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)

	last = [(4, 'week'), (5, 'month')]
	last_urls = [filter_urls(urls, i) for i in last]

	ranges = [(5, 'month', 4, 'week'), (7, 'year', 1, 'day')]
	ranges_urls = [filter_urls(urls, i[:2], i[2:]) for i in ranges]

	template = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	short = 'http://goo.gl/fDwgtb'
	expanded = 'http://python.org/'
	url = history.Url(short, expanded, datetime.datetime.now())

	short_dummies = ['a', 'b', 'c', 'd', 'e']
	dummies = [
		history.Url('a', 'aa', datetime.datetime(1997, 7, 26)),
		history.Url('b', 'bb', datetime.datetime(2009, 5, 23)),
		history.Url('c', 'cc', datetime.datetime(2012, 12, 21)),
		history.Url('d', 'dd', datetime.datetime(2015, 6, 3)),
		history.Url('e', 'ee', datetime.datetime(2015, 9, 15))
	]


	return Fixture(history,
				   urls,
				   last,
				   last_urls,
				   ranges,
				   ranges_urls,
				   template,
				   url,
				   dummies,
				   short_dummies)


def test_initializes_well(fixture):
	assert hasattr(fixture.history, 'raw')
	assert hasattr(fixture.history, 'delta')
	assert isinstance(fixture.history.delta, dict)
	values = fixture.history.delta.values()
	assert all(isinstance(i, datetime.timedelta) for i in values)

def test_request_works(fixture):
	expected = request()
	result = fixture.history.request()

	assert result == expected


def test_lineify_picks_short_if_pretty_false(fixture):
	result = fixture.history.lineify(fixture.url, False, False, False)

	assert result == fixture.url.short


def test_lineify_prettifies_if_pretty_true(fixture):
	result = fixture.history.lineify(fixture.url, False, False, True)
	expected = fixture.template.format(fixture.url.short)

	assert result == expected


def test_lineify_returns_only_expanded_if_expanded_true(fixture):
	result = fixture.history.lineify(fixture.url, True, False, False)

	assert result == fixture.url.long


def test_lineify_returns_both_if_both_true(fixture):
	result = fixture.history.lineify(fixture.url, False, True, False)
	expected = '{0} => {1}'.format(fixture.url.short, fixture.url.long)

	assert result == expected


def test_listify_works(fixture):
	result = fixture.history.listify([fixture.url],
									 None,
									 True,
									 False,
									 False)
	expected = [fixture.url.long]

	assert result == expected

def test_listify_limits_well(fixture):
	urls = [fixture.url] + fixture.dummies
	result = fixture.history.listify(urls,
									 1,
									 False,
									 False,
									 False)
	expected = [fixture.url.short]

	assert result == expected


def test_filter_works(fixture):
	begin = datetime.datetime(2008, 1, 3)
	end = datetime.datetime(2013, 4, 7)
	result = fixture.history.filter(fixture.dummies, begin, end)
	expected = fixture.dummies[1:3]

	assert result == expected


def test_get_date_works(fixture):
	now = datetime.datetime.now()
	result = fixture.history.get_date((1, 'minute'), now)
	expected = now - datetime.timedelta(minutes=1)

	assert result == expected


def test_get_date_works_works_if_endswith_s(fixture):
	now = datetime.datetime.now()
	result = fixture.history.get_date((1, 'minutes'), now)
	expected = now - datetime.timedelta(minutes=1)

	assert result == expected


def test_get_boundaries_works(fixture):
	base = datetime.datetime.now()
	result = fixture.history.get_boundaries(fixture.ranges[0], base)
	expected = (timestamp(fixture.ranges[0][:2], base),
				timestamp(fixture.ranges[0][2:], base))

	assert result == expected


def test_get_boundaries_throws_for_invalid_range(fixture):
	base = datetime.datetime.now()
	invalid_range = (4, 'days', 10, 'weeks')

	with pytest.raises(lnk.errors.UsageError):
		fixture.history.get_boundaries(invalid_range, base)


def test_last_works_for_single_range(fixture):
	result = fixture.history.last(fixture.dummies,
								  [fixture.last[1]],
								  None,
								  False,
								  False,
								  False)
	expected = fixture.short_dummies[3:]

	assert result == expected

def test_last_works_for_many_ranges(fixture):
	result = fixture.history.last(fixture.dummies,
								  fixture.last,
								  None,
								  False,
								  False,
								  False)
	expected = [fixture.short_dummies[4]]
	expected += fixture.short_dummies[3:]

	assert result == expected

def test_ranges_works_for_single_range(fixture):
	result = fixture.history.ranges(fixture.dummies,
									[fixture.ranges[0]],
									None,
									False,
									False,
									False)
	expected = [fixture.short_dummies[3]]

	assert result == expected

def test_ranges_works_for_many_ranges(fixture):
	result = fixture.history.ranges(fixture.dummies,
									fixture.ranges,
									None,
									False,
									False,
									False)
	expected = [fixture.dummies[3].short]
	expected += fixture.short_dummies[1:]

	assert result == expected


def test_forever_works(fixture):
	result = fixture.history.forever(fixture.dummies,
									 None,
									 False,
									 False,
									 False)
	expected = fixture.short_dummies

	assert result == expected


def test_pretty_works_for_forever(fixture):
	result = fixture.history.forever(fixture.dummies,
									 None,
									 False,
									 False,
									 True)
	expected = ['Since forever:']
	expected += [fixture.template.format(i) for i in fixture.short_dummies]

	assert result == expected + ['']

def test_pretty_works_for_last(fixture):
	result = fixture.history.last(fixture.all_urls,
								  fixture.last,
								  None,
								  False,
								  False,
								  True)
	expected = []
	for timespan, urls in zip(fixture.last, fixture.last_urls):
		span = '{0} '.format(timespan[0]) if timespan[0] > 1 else ''
		header = 'Last {0}{1}:'.format(span, timespan[1])
		if not urls:
			header += ' None'
		expected.append(header)
		for item in urls:
			expected.append(fixture.template.format(item.short))
		expected.append('')

	assert result == expected

def test_pretty_works_for_ranges(fixture):
	result = fixture.history.ranges(fixture.all_urls,
									fixture.ranges,
									None,
									False,
									False,
									True)
	expected = []
	for timespan, urls in zip(fixture.ranges, fixture.ranges_urls):
		header = 'Between {0} {1} '.format(timespan[0], timespan[1])
		header += 'and {0} {1} ago:'.format(timespan[2], timespan[3])
		if not urls:
			header += ' None'
		expected.append(header)
		for item in urls:
			expected.append(fixture.template.format(item.short))
		expected.append('')

	assert result == expected


def test_ranges_header_removes_unit_if_both_equal(fixture):
	time_range = (7, 'days', 2, 'days')
	result = fixture.history.ranges_header(time_range, True)
	expected = 'Between 7 and 2 days ago:'

	assert result == expected


def test_ranges_handles_empty_results_well(fixture):
	timespan = fixture.ranges[0]
	result = fixture.history.ranges([],
									[timespan],
									None,
									False,
									False,
									True)
	header = 'Between {0} {1} '.format(timespan[0], timespan[1])
	header += 'and {0} {1} ago:'.format(timespan[2], timespan[3])
	header += ' None'
	expected = [header, '']

	assert sorted(result) == sorted(expected)


def test_fetch_works_only_for_forever(fixture):
	result = fixture.history.fetch(None, None, True, None, False, False, False)
	expected = [i.short for i in fixture.all_urls]

	assert sorted(result) == sorted(expected)


def test_fetch_removes_last_line(fixture):
	result = fixture.history.fetch(None, None, True, None, False, False, True)
	expected = len(['Since forever:'] + [i.short for i in fixture.all_urls])

	assert len(result) == expected
	assert result[-1] != ''


def test_fetch_works_for_all_ranges(fixture):
	result = fixture.history.fetch(fixture.last,
								   fixture.ranges,
								   True,
								   None,
								   False,
								   False,
								   False)
	expected = [fixture.all_urls] + fixture.ranges_urls + fixture.last_urls
	expected = [j.short for i in expected for j in i]

	assert len(result) == len(expected)
	assert sorted(result) == sorted(expected)


def test_fetch_limits_well(fixture):
	result = fixture.history.fetch(None, None, True, 3, False, False, False)
	expected = [i.short for i in fixture.all_urls[:3]]

	assert len(result) <= 3
	assert sorted(result) == sorted(expected)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Link-history-retrieval for the bit.ly client."""

from __future__ import unicode_literals

import click
import time

import lnk.beauty
import lnk.bitly.link

from lnk.bitly.command import Command

def echo(*args):
	"""
	Executes a history command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 History instance's fetch() method.
	"""
	click.echo(History().fetch(*args))

class History(Command):
	"""
	Class to retrieve bit.ly link history for a user.

	The sole purpose of this class is to fetch and return a list of
	urls the user has shortened using bit.ly in the past. This list
	may, of course, be properly prettified if necessary. A nice feature
	is that if the history is fetched with the 'plain' flag set to true
	its output can be piped into other lnk commands, such as 'stat' or 'info'.

	Attributes:
		raw (bool): Whether to prettify the output or
					return it raw, for internal use.
		link (bit.ly.link.Link): A link instance to expand short links.
		seconds (dict): A mapping between string representations of time
						and equivalent numbers of seconds for each unit.
	"""
	def __init__(self, raw=False):
		super(History, self).__init__('history')
		self.raw = raw
		self.link = lnk.bitly.link.Link(raw=True)
		self.seconds = {
			'minute': 60, 
			'hour': 3600, 
			'day': 86400,
			'week': 604800, 
			'month': 2629740,
			'year': 31556900
		}

	def fetch(self, last, ranges, forever, limit, expanded, both, pretty):
		"""
		Fetches the link history.

		Arguments:
			last (tuple): A tuple of time_points to retrieve history for, without
						  an upper bound (i.e. all links after a given time_point,
						  up to now).
			ranges (tuple): A tuple of time_point-ranges to retrieve history for
							(with a lowe and upper bound).
			forever (bool): Whether or not to include data for all
							time_points (since forever).
			limit (int): A limit on the number of datapoints fetched per time_point.
			expanded (bool): Whether or not to show expanded rather
							 than short links.
			both (bool): Whether or not to show both short and expanded links
						 (takes precedence over 'expanded').
			pretty (bool): Whether or not to prettify the output.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is true,
			else a boxified, pretty string if the pretty flag is set,
			else the same plain list as for the first case, but joined to
			a string.
		"""
		self.parameters['limit'] = limit
		result = []
		if forever:
			result += self.forever(expanded, both, pretty)
		if ranges:
			result += self.ranges(set(ranges), expanded, both, pretty)
		if last:
			result += self.last(set(last), expanded, both, pretty)

		# Remove last empty line
		if pretty:
			del result[-1]

		if self.raw:
			return result
		return lnk.beauty.boxify([result]) if pretty else '\n'.join(result)

	def forever(self, expanded, both, pretty):
		"""
		Fetches and formats history since forever.

		Arguments:
			expanded (bool): Whether to show exapanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header ('Since forever:') plus
			an empty line for padding.
		"""
		lines = []
		for url in self.request():
			line = self.lineify(url, expanded, both, pretty)
			lines.append(line)

		return ['Since forever:'] + lines + [''] if pretty else lines

	def ranges(self, ranges, expanded, both, pretty):
		"""
		Fetches and formats history for certain time ranges.

		Arguments:
			ranges (tuple): The time-ranges of schema (span1, unit1, span2, unit2).
			expanded (bool): Whether to show expanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header (e.g. 'Between 7 months and
			5 days ago:'), plus an empty line for padding.
		"""
		lines = []
		for time_point in ranges:
			before = time_point[2:]
			after = time_point[:2]
			parameters = self.parse_time(after, before)
			urls = self.request(parameters)
			if pretty:
				header = self.ranges_header(after, before, urls)
				lines.append(header)
			for url in urls:
				line = self.lineify(url, expanded, both, pretty)
				lines.append(line)
			if pretty:
				lines.append('')

		return lines

	def last(self, last, expanded, both, pretty):
		"""
		Fetches and formats history after a given time-point.

		Arguments:
			last (tuple): The open-ended time-ranges of schema (span, unit).
			expanded (bool): Whether to show exapanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header (e.g. 'Last 4 weeks:'),
			plus an empty line for padding.
		"""
		lines = []
		for time_point in last:
			parameters = self.parse_time(time_point)
			urls = self.request(parameters)
			if pretty:
				header = self.last_header(time_point, urls)
				lines.append(header)
			for url in urls:
				line = self.lineify(url, expanded, both, pretty)
				lines.append(line)
			if pretty:
				lines.append('')

		return lines

	def ranges_header(self, after, before, urls):
		"""
		Returns a header for a time-range.

		Arguments:
			after (tuple): The lower-bound time-point.
			before (tuple): The upper-bound time-point.
			urls (list|tuple): The urls, to check if empty.

		Returns:
			A header of the schema 'Between <span1> [<unit1>] and <span2>
			<unit2>', where <unit1> is only included if it not equal to
			<unit2>, i.e. 'Between 3 days and 2 days ago' is transformed to
			'Between 3 and 2 days ago'. If the urls are empty, ' None' is
			appended to the header.
		"""
		first_unit = ''
		if before[1] != after[1]:
			first_unit = '{0} '.format(after[1])
		header = 'Between {0} {1}'.format(after[0], first_unit)
		header += 'and {0} {1} ago:'.format(before[0], before[1])

		return header if urls else header + ' None'

	def last_header(self, time_point, urls):
		"""
		Returns a header for an open-ended time-range.

		Arguments:
			time_point (tuple): The lower-bound time-point.
			urls (list|tuple): The urls, to check if empty.

		Returns:
			A header of the schema 'Last [<span>] <unit>'. The span is
			only shown if it is greater 1, such that 'Last 1 day' will be
			displayed as 'Last day'. If the urls are empty, ' None' is 
			appended to the header.
		"""
		span = '{0} '.format(time_point[0]) if time_point[0] > 1 else ''
		header = 'Last {0}{1}:'.format(span, time_point[1])

		return header if urls else header + ' None'

	def parse_time(self, after=None, before=None, base=None):
		"""
		Parses one or two time-points for an HTTP request.

		The time-range can have a lower-bound and an upper-bound,
		but does not need to have either.

		Arguments:
			after (tuple): The lower-bound time-point.
			before (tuple): The upper-bound time-point.

		Returns:
			Parameters ready for an HTTP request to the bit.ly API,
			as a dictionary (dict).
		"""
		if before:
			before = self.timestamp(before, base)
		parameters = {
			'created_before': before,
			'created_after': self.timestamp(after, base)
		}

		return parameters

	def timestamp(self, time_point, base=None):
		"""
		Returns a UNIX timestamp for a given time-point.

		The timestamp is base timestamp (either supplied to the function
		call or time.time()) minus the time_point, parsed into seconds.

		Arguments:
			time_point (tuple): A time_point tuple of the schema (span, unit).
			base (int): The base time-value from which to subtract the offset,
						defaults to time.time().

		Returns:
			A UNIX timestamp, in seconds, as an integer (!).
		"""
		span = time_point[0]
		unit = time_point[1]
		if unit.endswith('s'):
			unit = unit[:-1]
		offset = span * self.seconds[unit]
		base = base or time.time()

		return int(base - offset)

	def lineify(self, url, expanded, both, pretty):
		"""
		Returns an output-ready line for a given url.

		The line may include only the short url, only the long/expanded
		url or both, depending on the parameters passed.

		Arguments:
			url (str): The short url.
			expanded (bool): Whether or not to show the expanded link.
			both (bool): Whether or not to show both the short and the
						 expanded link (takes precedence over 'expanded').
			pretty (bool): Whether or not to make the output pretty.

		Returns:
			The short url if neither 'expanded' nor 'both' is set,
			both the short and the expanded url, in the schema '<short> =>
			<expanded>' if 'both' is set, else only the expanded link if
			'expanded' is set. If 'pretty' is True, the line is returned
			in a formatted fashion (with color), else in plain format.
		"""
		if both:
			expanded = self.link.get_long(url)
			url = '{0} => {1}'.format(url, expanded)
		elif expanded:
			url = self.link.get_long(url)

		if pretty:
			url = self.list_item.format(url)

		return url

	def request(self, parameters=None):
		"""
		Requests the link-history from the bit.ly API.

		Arguments:
			parameters (dict): Parameters to pass along with the request 
							   (like created_after/created_before).

		Returns:
			A list of links.
		"""
		response = self.get(self.endpoints['history'], parameters)
		response = self.verify(response, 'retrieve history')

		return [i['link'] for i in response['link_history']]



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

import os
import json

import lnk.errors

def get(which, key):
	"""
	Convenience method to retrieve only one
	key without having to instantiate a Manager object.
	"""
	return Manager(which)[key]

class Manager(object):

	def __init__(self, which=None, write=False):

		path = os.path.abspath(__file__)

		parent = os.path.dirname(os.path.dirname(os.path.dirname(path)))

		self.path = os.path.join(parent, 'config')

		self.write_upon_exit = write
		self.which = which
		self.file = None
		self.config = self.open(which) if which else None

	def open(self, which):
		self.file = os.path.join(self.path, '{0}.json'.format(which))
		with open(self.file) as source:
			self.config = json.load(source)
		return self.config

	def close(self):
		self.assert_open()
		self.file = self.config = None

	def write(self):
		self.assert_open()
		with open(self.file, 'wt') as destination:
			json.dump(self.config, destination, indent=4)

	def assert_open(self):
		if not self.file:
			what = 'No configuration file was ever opened!'
			raise lnk.errors.InternalError(what)

	@property
	def keys(self):
		return self.config.keys()

	@property
	def values(self):
		return self.config.values()

	@property
	def items(self):
		return self.config.items()

	def __getitem__(self, key):
		if key not in self.config:
			raise lnk.errors.InvalidKeyError("Key '{0}' not found.".format(key))
		return self.config[key]

	def __setitem__(self, key, value):
		if key not in self.config:
			raise lnk.errors.InvalidKeyError("Key '{0}' not found.".format(key))
		self.config[key] = value

	def __enter__(self):
		return self

	def __exit__(self, type, value, traceback):
		if self.write_upon_exit:
			self.write()
		self.close()



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Contains the base-class for all bit.ly commands."""

import lnk.config
import lnk.errors

from lnk.abstract import AbstractCommand

class Command(AbstractCommand):
	"""
	Base-class for all tinyurl commands.

	Configures the AbstractCommand base class for all commands in the
	entire application, which needs information about the service being
	used. Moreover sets up the necessary parameters needed for any request
	to the bit.ly API (the api-key, the response-format and the provider).

	Attributes:
		parameters (dict): The necessary parameters for any request to the
						   tinyurl API.
	"""
	def __init__(self, which):
		super(Command, self).__init__('tinyurl', which)
		with lnk.config.Manager('tinyurl') as manager:
			self.parameters = {'apikey': manager['key']}
		self.parameters['format'] = 'json'
		self.parameters['provider'] = 'tinyurl_com'

	@staticmethod
	def verify(response, what):
		"""
		Verifies an HTTP-response from the tinyurl API.

		Overrides the 'pure-virtual' (i.e. not-implemented) base method
		from AbstractCommand. If the verification finds no faults in the
		response, the data is returned.

		Arguments:
			response (requests.Response): The HTTP response to a request
										  to the bit.ly API.
			what (str): A human-readable string representing what the request
						was for, such that if there is an error in the response,
						an errors.HTTPError or errors.APIError is raised with
						the message 'Could not <what>.'

		Returns:
			The actual data of the response, if no fault was found.

		Raises:
			errors.HTTPError: If it was found that there was an HTTP-related
							  exception, such as a faulty URL or other badness.
		"""
		if not str(response.status_code).startswith('2'):
			raise lnk.errors.HTTPError('Could not {0}!'.format(what),
								   response.status_code,
								   response.reason)
		data = response.json()
		if data['state'] == 'error':
			what = 'Could not {0}!'.format(what)
			raise lnk.errors.APIError(what)

		return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ecstasy
import pyperclip
import pytest
import requests
import threading

from collections import namedtuple

import tests.paths
import lnk.tinyurl.link

def shorten(url):
	response = requests.get('http://tiny-url.info/api/v1/create',
							params=dict(apikey='0BFA4A7B5BDD5BE7780C',
									 	format='json',
									 	provider='tinyurl_com',
									 	url=url))
	data = response.json()

	return data['shorturl']

@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', ['link', 'long', 'short', 'formatted'])
	link = lnk.tinyurl.link.Link(raw=True)
	url = 'https://www.github.com/goldsborough/lnk'
	short = shorten(url)
	formatted = ecstasy.beautify('<{0}>'.format(short), ecstasy.Style.Bold)

	return Fixture(link, url, short, formatted)

def test_copy_copies_to_clipboard_if_copy_true(fixture):
	fixture.link.copy(True, fixture.short)

	assert pyperclip.paste() == fixture.short

def test_copy_copies_only_first_url(fixture):
	assert fixture.link.already_copied

	fixture.link.copy(True, 'a')
	fixture.link.copy(True, 'b')
	fixture.link.copy(True, 'c')

	assert pyperclip.paste() == fixture.short

def test_copy_copies_to_clipboard_if_copy_false(fixture):
	pyperclip.copy('original')
	fixture.link.copy(False, fixture.short)

	assert pyperclip.paste() == 'original'

def test_copy_makes_copied_url_bold(fixture):
	fixture.link.already_copied = False
	returned_url = fixture.link.copy(True, fixture.short)

	assert returned_url == fixture.formatted

def test_request_shortens_well(fixture):
	result = fixture.link.request(fixture.long)

	assert result == fixture.short

def test_shorten_formats_well(fixture):
	result = []
	fixture.link.shorten(result, False, True, True, fixture.long)
	expected = '{0} => {1}'.format(fixture.long, fixture.short)

	assert result[0] == expected

def test_shorten_warns_about_url_without_protocol(fixture, capsys):
	fixture.link.shorten([], False, False, False, 'google.com')
	out = capsys.readouterr()

	assert out
	assert out[0].startswith("\aWarning: Prepending 'http://' to")


def test_fetch_works_for_single_url(fixture):
	result = fixture.link.fetch(False, True, [fixture.long], False)

	assert result == [fixture.short]

def test_fetch_works_for_many_urls(fixture):
	urls = ['http://facebook.com', 'http://google.com', 'http://python.org']
	result = set(fixture.link.fetch(False, True, urls, False))

	expected = set()
	threads = []
	for url in urls:
		thread = threading.Thread(target=lambda u: expected.add(shorten(u)),
								  args=(url,))
		thread.daemon = True
		thread.start()
		threads.append(thread)

	for thread in threads:
		thread.join(timeout=10)

	assert result == expected

def test_fetch_correct_output_if_raw_false_pretty_false(fixture):
	fixture.link.raw = False
	urls = [fixture.long, 'http://python.org']
	result = fixture.link.fetch(False, True, urls, False)
	expected = '\n'.join([fixture.short, shorten(urls[1])])

	return result == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import ecstasy
import os
import pytest
import requests
import time

from collections import namedtuple

import tests.paths
import lnk.bitly.history

VERSION = 3
API = 'https://api-ssl.bitly.com/v{0}'.format(VERSION)
with open(os.path.join(tests.paths.TEST_PATH, 'bitly', 'token')) as source:
	ACCESS_TOKEN = source.read()


def timestamp(time_range):
	seconds = {
		"minute": 60, 
		"hour": 3600, 
		"day": 86400,
		"week": 604800, 
		"month": 2629740,
		"year": 31556900
	}
	offset = time_range[0] * seconds[time_range[1]]

	return int(time.time() - offset)

def request_history(start=None, end=None, limit=None):
	if start:
		start = timestamp(start)
	if end:
		end = timestamp(end)
	response = requests.get('{0}/user/link_history'.format(API),
							params=dict(access_token=ACCESS_TOKEN,
										created_after=start,
										created_before=end,
										limit=limit))
	data = response.json()['data']

	return [i['link'] for i in data['link_history']]


def request_expansion(url):
	response = requests.get('{0}/expand'.format(API),
							params=dict(access_token=ACCESS_TOKEN,
										shortUrl=url))
	data = response.json()['data']

	return data['expand'][0]['long_url']


@pytest.fixture(scope='module')
def fixture():
	
	Fixture = namedtuple('Fixture', [
		'history',
		'forever_data',
		'last',
		'last_data',
		'ranges',
		'ranges_data',
		'template',
		'url',
		'expanded'
		])

	history = lnk.bitly.history.History(raw=True)
	forever_data = request_history()

	last = [(4, 'week'), (5, 'month')]
	last_data = [request_history(i) for i in last]

	ranges = [(5, 'month', 4, 'day'), (7, 'year', 1, 'day')]
	ranges_data = [request_history(i[:2], i[2:]) for i in ranges]

	template = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	url = 'http://bit.ly/1OQM9nA'
	expanded = request_expansion(url)

	return Fixture(history,
				   forever_data,
				   last,
				   last_data,
				   ranges,
				   ranges_data,
				   template,
				   url,
				   expanded)


def test_initializes_well(fixture):
	assert hasattr(fixture.history, 'raw')
	assert hasattr(fixture.history, 'link')
	assert hasattr(fixture.history, 'seconds')
	assert isinstance(fixture.history.seconds, dict)

def test_request_works(fixture):
	expected = request_history()
	result = fixture.history.request()

	print(result, expected)

	assert sorted(result) == sorted(expected)


def test_lineify_does_nothing_if_pretty_false(fixture):
	result = fixture.history.lineify('cat', False, False, False)

	assert result == 'cat'


def test_lineify_prettifies_if_pretty_true(fixture):
	result = fixture.history.lineify('cat', False, False, True)
	expected = fixture.template.format('cat')

	assert result == expected


def test_lineify_returns_only_expanded_if_expanded_true(fixture):
	result = fixture.history.lineify(fixture.url, True, False, False)

	assert result == fixture.expanded


def test_lineify_returns_both_if_both_true(fixture):
	result = fixture.history.lineify(fixture.url, False, True, False)
	expected = '{0} => {1}'.format(fixture.url, fixture.expanded)

	assert result == expected


def test_timestamp_works(fixture):
	now = time.time()
	result = fixture.history.timestamp((1, 'minute'), now)
	expected = int(now - 60)

	assert result == expected


def test_timestamp_works_if_endswith_s(fixture):
	now = time.time()
	result = fixture.history.timestamp((1, 'minutes'), now)
	expected = int(now - 60)

	assert result == expected	


def test_parse_time_works_without_upper_bound(fixture):
	now = time.time()
	result = fixture.history.parse_time((1, 'minute'), base=now)
	expected = int(now - 60)

	assert result['created_after'] == expected
	assert result['created_before'] is None


def test_parse_time_works_with_upper_bound(fixture):
	now = int(time.time())
	result = fixture.history.parse_time((2, 'minute'), (1, 'minute'), now)

	assert result['created_after'] == now - 120
	assert result['created_before'] == now - 60


def test_last_works_for_single_range(fixture):
	result = fixture.history.last([fixture.last[0]], False, False, False)
	expected = fixture.last_data[0]

	assert result == expected

def test_last_works_for_many_ranges(fixture):
	result = fixture.history.last(fixture.last, False, False, False)
	expected = fixture.last_data[0] + fixture.last_data[1]

	assert result == expected

def test_ranges_works_for_single_range(fixture):
	result = fixture.history.ranges([fixture.ranges[0]], False, False, False)
	expected = fixture.ranges_data[0]

	assert result == expected


def test_ranges_works_for_many_ranges(fixture):
	result = fixture.history.ranges(fixture.ranges, False, False, False)
	expected = [j for i in fixture.ranges_data for j in i]

	assert sorted(result) == sorted(expected)


def test_forever_works(fixture):
	result = fixture.history.forever(False, False, False)
	expected = fixture.forever_data

	assert result == expected


def test_pretty_works_for_forever(fixture):
	result = fixture.history.forever(False, False, True)
	expected = ['Since forever:']
	expected += [fixture.template.format(i) for i in fixture.forever_data]

	assert result == expected + ['']

def test_pretty_works_for_last(fixture):
	result = fixture.history.last(fixture.last, False, False, True)
	expected = []
	for time_point, data in zip(fixture.last, fixture.last_data):
		header = 'Last {0} {1}:'.format(time_point[0], time_point[1])
		if not data:
			header += ' None'
		expected.append(header)
		for item in data:
			expected.append(fixture.template.format(item))
		expected.append('')

	assert sorted(result) == sorted(expected)

def test_ranges_header_removes_unit_if_both_equal(fixture):
	after = (7, 'days')
	before = (2, 'days')
	result = fixture.history.ranges_header(after, before, True)
	expected = 'Between 7 and 2 days ago:'

	assert result == expected

def test_pretty_works_for_ranges(fixture):
	result = fixture.history.ranges(fixture.ranges, False, False, True)
	expected = []
	for time_point, data in zip(fixture.ranges, fixture.ranges_data):
		header = 'Between {0} {1} '.format(time_point[0], time_point[1])
		header += 'and {0} {1} ago:'.format(time_point[2], time_point[3])
		if not data:
			header += ' None'
		expected.append(header)
		for item in data:
			expected.append(fixture.template.format(item))
		expected.append('')

	assert sorted(result) == sorted(expected)


def test_fetch_works_only_for_forever(fixture):
	result = fixture.history.fetch(None, None, True, None, False, False, False)
	expected = fixture.forever_data

	assert result == expected


def test_fetch_removes_last_line(fixture):
	result = fixture.history.fetch(None, None, True, None, False, False, True)
	expected = len(['Since forever:'] + fixture.forever_data)

	assert len(result) == expected
	assert result[-1] != ''


def test_fetch_works_for_all_ranges(fixture):
	result = fixture.history.fetch(fixture.last,
								   fixture.ranges,
								   True,
								   None,
								   False,
								   False,
								   False)
	expected = [fixture.forever_data] + fixture.last_data + fixture.ranges_data
	expected = [j for i in expected for j in i]

	assert sorted(result) == sorted(expected)


def test_fetch_limits_well(fixture):
	result = fixture.history.fetch(None, None, True, 3, False, False, False)
	expected = fixture.forever_data[:3]

	assert len(result) <= 3
	assert result == expected
